//! A very inaccurate gameboy emulator
#![feature(decl_macro)]
#![allow(clippy::unit_arg)]

/// A gameboy has:
/// - A 16-bit memory bus
/// - A CPU
/// - A picture processing unit

pub mod error {
    use crate::cpu::disasm::Insn;
    use std::fmt::Display;

    #[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord)]
    pub struct Error {
        kind: ErrorKind,
        // any other information about an error
    }

    #[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord)]
    pub enum ErrorKind {
        InvalidInstruction(u8),
        InvalidAddress(u16),
        UnimplementedInsn(Insn),
        Unimplemented(u8),
        HitBreak(u16, u8),
        Todo,
    }
    impl From<ErrorKind> for Error {
        fn from(value: ErrorKind) -> Self {
            Self { kind: value }
        }
    }
    impl<T> From<Error> for Result<T, Error> {
        fn from(value: Error) -> Self {
            Err(value)
        }
    }
    impl Display for Error {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            self.kind.fmt(f)
        }
    }
    impl Display for ErrorKind {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                ErrorKind::InvalidInstruction(i) => write!(f, "Invalid instruction: {i}"),
                ErrorKind::InvalidAddress(addr) => write!(f, "Invalid address: {addr:04x}"),
                ErrorKind::UnimplementedInsn(i) => write!(f, "Unimplemented instruction: {i}"),
                ErrorKind::Unimplemented(i) => write!(f, "Unimplemented instruction: {i}"),
                ErrorKind::HitBreak(addr, m) => {
                    write!(f, "Hit breakpoint {addr:x} (Took {m} cycles)")
                }
                ErrorKind::Todo => write!(f, "Not yet implemented"),
            }
        }
    }
}

pub mod memory;

pub mod graphics {
    use bitfield_struct::bitfield;

    use crate::memory::io::BusIO;
    use std::mem::size_of;

    /// The number of OAM entries supported by the PPU (on real hardware, 40)
    pub const NUM_OAM_ENTRIES: usize = 40;

    #[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
    pub enum Mode {
        HBlank,
        VBlank,
        OAMScan,
        Drawing,
    }
    impl Mode {
        const fn from_bits(value: u8) -> Self {
            match value & 3 {
                0b00 => Self::HBlank,
                0b01 => Self::VBlank,
                0b10 => Self::OAMScan,
                0b11 => Self::Drawing,
                _ => unreachable!(),
            }
        }
        const fn into_bits(self) -> u8 {
            self as _
        }
    }

    type OAMTable = [OAMEntry; NUM_OAM_ENTRIES];

    #[derive(Clone, Debug, PartialEq, PartialOrd)]
    pub struct PPU {
        lcdc: LcdC, // LCD Control
        stat: Stat, // LCD Status
        scy: u8,    // viewport y coordinate
        scx: u8,    // viewport x coordinate
        /// Current line. Read-only.
        ly: u8, // holds values in range 0..=153 (144..=153 = vblank)
        /// Line Y for Comparison. Triggers interrupt if enabled in `stat`
        lyc: u8, // line y compare, triggers interrupt if enabled

        oam: OAMTable,
    }

    impl PPU {
        pub fn oam(&self) -> &[u8] {
            let data = self.oam.as_slice().as_ptr().cast();
            let len = size_of::<OAMTable>();
            // SAFETY:
            // [OAMEntry; NUM_OAM_ENTRIES] is sizeof<OAM_ENTRY> * NUM_OAM_ENTRIES long,
            // and already aligned
            unsafe { core::slice::from_raw_parts(data, len) }
        }
        pub fn oam_mut(&mut self) -> &mut [u8] {
            let data = self.oam.as_mut_slice().as_mut_ptr().cast();
            let len = size_of::<OAMTable>();
            // SAFETY: see above
            unsafe { core::slice::from_raw_parts_mut(data, len) }
        }
        /// Gets the current PPU operating mode
        pub fn mode(&self) -> Mode {
            self.stat.mode()
        }
        pub fn set_mode(&mut self, mode: Mode) {
            self.stat.set_mode(mode);
        }
        /// Gets the STAT interrupt selection config.
        ///
        /// | Bit | Name           |
        /// |:---:|:---------------|
        /// | `0` | H Blank        |
        /// | `1` | V Blank        |
        /// | `2` | OAM Scan       |
        /// | `3` | LY Coincidence |
        pub fn stat_interrupt_selection(&self) -> u8 {
            (self.stat.into_bits() & 0b1111000) >> 3
        }
        /// Gets the LY coincidence flag. Should be true when LY = LYC
        pub fn ly_coincidence(&self) -> bool {
            self.stat.lyc_hit()
        }
    }

    impl Default for PPU {
        fn default() -> Self {
            Self {
                lcdc: Default::default(),
                stat: Default::default(),
                scy: Default::default(),
                scx: Default::default(),
                ly: Default::default(),
                lyc: Default::default(),
                oam: [Default::default(); NUM_OAM_ENTRIES],
            }
        }
    }
    impl BusIO for PPU {
        fn read(&self, addr: usize) -> Option<u8> {
            match addr {
                0xfe00..=0xfe9f => self.oam().read(addr - 0xfe00),
                0xff40 => Some(self.lcdc.into_bits()),
                0xff41 => Some(self.stat.into_bits()),
                0xff42 => Some(self.scy),
                0xff43 => Some(self.scx),
                0xff44 => Some(self.ly),
                0xff45 => Some(self.lyc),
                _ => None,
            }
        }

        fn write(&mut self, addr: usize, data: u8) -> Option<()> {
            match addr {
                0xfe00..=0xfe9f => self.oam_mut().write(addr - 0xfe00, data),
                0xff40 => Some(self.lcdc = LcdC::from_bits(data)),
                0xff41 => Some(self.stat.set_interrupts(Stat::from_bits(data).interrupts())),
                0xff42 => Some(self.scy = data),
                0xff43 => Some(self.scx = data),
                0xff44 => None,
                0xff45 => Some(self.lyc = data),
                _ => None,
            }
        }
    }

    #[repr(C)]
    #[derive(Clone, Copy, Debug, Default, PartialEq, Eq, PartialOrd, Ord, Hash)]
    pub struct OAMEntry {
        y: u8,         // y coordinate of the bottom of a 16px sprite
        x: u8,         // x coordinate of the right of an 8px sprite
        tid: u8,       // tile identifier
        attr: OAMAttr, // OAM Attributes
    }

    #[bitfield(u8)]
    /// OAM Attributes. Read-writable by CPU, read-only by PPU.
    #[derive(PartialEq, Eq, PartialOrd, Ord, Hash)]
    pub struct OAMAttr {
        #[bits(3)]
        /// Which palette (OBP 0-7) does this object use in CGB mode?
        cgb_palette: u8,
        #[bits(1)]
        /// Which VRAM bank (0-1) contains this object's tile?
        cgb_bank: u8,
        #[bits(1)]
        /// Which palette (OBP 0-1) does this object use in DMG mode?
        dmg_palette: u8,
        /// Is this sprite flipped horizontally?
        x_flip: bool,
        /// Is this sprite flipped vertically?
        y_flip: bool,
        /// Is this sprite drawn below the window and background layers?
        priority: bool,
    }

    #[bitfield(u8)]
    /// LCD main control register
    #[derive(PartialEq, Eq, PartialOrd, Ord, Hash)]
    pub struct LcdC {
        /// In DMG mode, disables Window and Background layers.
        /// In CGB mode, causes objects to render on top of Window and Background layers
        bg_window_enable_or_priority: bool,
        /// Controls whether objects are rendered or not
        obj_enable: bool,
        /// Controls the height of objects: false = 8px tall, true = 16px tall
        obj_size: bool,
        /// Controls bit 10 of the backgroud tile map address (false, true) => (0x9800, 0x9c00)
        bg_tile_map_area: bool,
        /// Controls which addressing mode to use when picking tiles for the window and background layers
        bg_window_tile_data_area: bool,
        /// Enables or disables the window
        window_enable: bool,
        /// Controls bit 10 of the window tile map address (false, true) => (0x9800, 0x9c00)
        window_tile_map_area: bool,
        /// Controls whether the LCD/PPU recieves power
        lcd_ppu_enable: bool,
    }

    #[bitfield(u8)]
    /// LCD Status and Interrupt Control register
    #[derive(PartialEq, Eq, PartialOrd, Ord, Hash)]
    pub struct Stat {
        #[bits(2)]
        /// The current PPU operating [Mode]
        mode: Mode,
        /// Set to true when LYC == LY. Read only.
        lyc_hit: bool,
        #[bits(4)]
        /// Configures PPU to raise IRQ when a condition is met
        interrupts: Interrupts,
        #[bits(default = true)]
        _reserved: bool,
    }

    #[bitfield(u8)]
    /// [Stat] register interrupt configuration
    pub struct Interrupts {
        /// Configures PPU to raise IRQ on VBlank start
        hblank: bool,
        /// Configures PPU to raise IRQ on VBlank start
        vblank: bool,
        /// Configures PPU to raise IRQ on OAM Scan start
        oam: bool,
        /// Configures PPU to raise IRQ on rising edge of [lyc_hit](Stat::lyc_hit)
        lyc: bool,
        #[bits(4)]
        _unused: (),
    }

    #[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
    pub enum DMGColor {
        White,
        LGray,
        DGray,
        Black,
    }

    impl DMGColor {
        pub const fn from_bits(value: u8) -> Self {
            match value & 3 {
                0 => DMGColor::White,
                1 => DMGColor::LGray,
                2 => DMGColor::DGray,
                3 => DMGColor::Black,
                _ => unreachable!(),
            }
        }
        pub const fn to_bits(self) -> u8 {
            self as _
        }
    }
}

pub mod cpu;

pub mod audio {
    use crate::memory::io::BusIO;
    use bitfield_struct::bitfield;

    #[derive(Debug, Default)]
    pub struct APU {
        wave: [u8; 0x10],
        amc: AMC,
        pan: SndPan,
        vol: Vol,
    }

    impl BusIO for APU {
        fn read(&self, addr: usize) -> Option<u8> {
            match addr {
                0xff10 => todo!("Channel 1 sweep"),
                0xff11 => todo!("Channel 1 length timer & duty cycle"),
                0xff12 => todo!("Channel 1 volume & envelope"),
                0xff13 => todo!("Channel 1 period low [write-only]"),
                0xff14 => todo!("Channel 1 period high & control [trigger, length enable, period]"),
                0xff24 => Some(self.vol.into_bits()),
                0xff25 => Some(self.pan.into_bits()),
                0xff26 => Some(self.amc.into_bits()),
                0xff30..=0xff3f => self.wave.read(addr - 0xff30),
                _ => None,
            }
        }

        fn write(&mut self, addr: usize, data: u8) -> Option<()> {
            #[allow(clippy::unit_arg)]
            match addr {
                0xff10..=0xff14 => todo!("Channel 1"),
                0xff16..=0xff19 => todo!("Channel 2"),
                0xff1a..=0xff1e => todo!("Channel 3"),
                0xff20..=0xff23 => todo!("Channel 4"),
                0xff24 => Some(self.vol = Vol::from_bits(data)),
                // Sound panning
                0xff25 => Some(self.pan = SndPan::from_bits(data)),
                // Audio Master Control
                0xff26 => Some(self.amc.set_power(data & 0x80 != 0)),
                // Wave table memory
                0xff30..=0xff3f => self.wave.write(addr - 0xff30, data),
                _ => None,
            }
        }
    }

    #[bitfield(u8)]
    /// Audio Master Control register
    pub struct AMC {
        /// Set when CH1 is playing. Read-only.
        ch1_on: bool,
        /// Set when CH2 is playing. Read-only.
        ch2_on: bool,
        /// Set when CH3 is playing. Read-only.
        ch3_on: bool,
        /// Set when CH4 is playing. Read-only.
        ch4_on: bool,
        #[bits(3)] // 3 bits are unused
        _reserved: (),
        /// If false, all channels are muted. Read-writable.
        power: bool,
    }

    #[bitfield(u8)]
    /// Sound Panning register. Setting `cNS` to true will send channel `N` to speaker `S`
    pub struct SndPan {
        c1r: bool,
        c2r: bool,
        c3r: bool,
        c4r: bool,
        c1l: bool,
        c2l: bool,
        c3l: bool,
        c4l: bool,
    }

    #[bitfield(u8)]
    /// Volume control and VIN panning register
    pub struct Vol {
        #[bits(3)]
        /// The volume of the right output
        right: u8,
        /// Whether VIN is sent to the right output
        vin_right: bool,
        #[bits(3)]
        /// The volume of the left output
        left: u8,
        /// Whether VIN is sent to the left output
        vin_left: bool,
    }
}

#[cfg(test)]
mod tests {
    use crate::memory::{io::*, Cart};
    const TEST_ROMS: &[&str] = &[
        // "roms/Legend of Zelda, The - Link's Awakening DX (USA, Europe) (Rev B) (SGB Enhanced).gbc",
        // "roms/Pokemon - Crystal Version (USA, Europe) (Rev A).gbc",
        // "roms/Pokemon Pinball (USA) (Rumble Version) (SGB Enhanced).gbc",
        // "roms/Pokemon - Red Version (USA, Europe) (SGB Enhanced).gb",
        "roms/Super Mario Land 2 - 6 Golden Coins (USA, Europe) (Rev B).gb",
        "roms/Super Mario Land (World) (Rev A).gb",
        "roms/Tetris (World) (Rev A).gb",
    ];
    #[test]
    /// Checksums the cart headers of every cart defined in [TEST_ROMS],
    /// and compares them against the checksum stored in the header
    fn cart_header_checksum() {
        for rom in TEST_ROMS {
            let rom = std::fs::read(rom).unwrap();
            let cart = Cart::from(rom);
            let mut cksum = 0u8;
            for index in 0x134..0x14d {
                cksum = cksum
                    .wrapping_sub(cart.read(index).unwrap())
                    .wrapping_sub(1)
            }
            assert_eq!(cksum, cart.headercksum().unwrap())
        }
    }
}
